using System;
using System.Collections;
using System.Xml;
using System.Drawing;
using System.Threading;

public delegate void NextGenerationCompletedHandler();

class GameOfLife
{
	// Members
	private Cell[,] boardBuffer1, boardBuffer2, publicBoard, backgroundBoard;
	private Size size;
	private bool doAutoMode;
	private int msPerGeneration;
	private int nrOfGeneration;
	
	// Members
	public event NextGenerationCompletedHandler NextGenerationCompleted;
	
	// C'tor
	public GameOfLife(Size size)
	{
		this.size = size;
		
		this.doAutoMode = false;
		this.msPerGeneration = 5000;
		this.nrOfGeneration = 0;
		
		this.CreateBoard(size);
	}
	
	// Getter
	public Cell[,] Board
	{
		get {return this.publicBoard;}
	}
	
	public Size Size
	{
		get {return this.size;}
	}
	
	public int NrOfGeneration
	{
		get {return this.nrOfGeneration;}
	}
	
	// Methods
	private void switchBuffer()
	{
		this.publicBoard = (this.publicBoard == this.boardBuffer1) ? this.boardBuffer2 : this.boardBuffer1;
		this.backgroundBoard = (this.backgroundBoard == this.boardBuffer1) ? this.boardBuffer2 : this.boardBuffer1;
	}

	private void CreateBoard(Size size)
	{
		this.size = size;
		this.boardBuffer1 = new Cell[size.Height, size.Width];
		this.boardBuffer2 = new Cell[size.Height, size.Width];

		for (int y = 0; y < size.Height; y++)
			for (int x = 0; x < size.Width; x++)
			{
				this.boardBuffer1[y, x] = new Cell();
				this.boardBuffer2[y, x] = new Cell();
			}

		this.publicBoard = boardBuffer1;
		this.backgroundBoard = boardBuffer2;
	}

	public void LoadPattern(string filename)
	{
		if(GOLXml.ValidateXml(filename))
		{
			XmlDocument xmlDoc = GOLXml.OpenXmlDoc(filename);
			Size s = GOLXml.GetBoardSize(xmlDoc);
			Point[] p = GOLXml.GetLivingCells(xmlDoc);
			
			this.CreateBoard(s);
			for (int i=0; i<p.Length; i++)
				this.backgroundBoard[p[i].Y, p[i].X].IsAlive = true;

			this.switchBuffer();

			if (this.NextGenerationCompleted != null)
				this.NextGenerationCompleted();
			
		}
	}
	
	public void SavePattern(string filename)
	{
		ArrayList a = new ArrayList();

		for (int y = 0; y < size.Height; y++)
			for (int x = 0; x < size.Width; x++)
				if (publicBoard[y, x].IsAlive)
					a.Add(new Point(x, y));

		GOLXml.SaveBoardState(this.size, a, filename);
	}
	
	public void AutoModeStart(int msPerGeneration)
	{
		this.msPerGeneration = msPerGeneration;
		this.doAutoMode = true;
		ThreadStart ts = new ThreadStart(this.AutoMode);
		Thread t = new Thread(ts);
		t.IsBackground = true;
		t.Start();
	}

	public void AutoModeStop()
	{
		this.doAutoMode = false;
	}
	
	private void AutoMode()
	{
		while(this.doAutoMode)
		{
			this.BuildNextGeneration();
			Thread.Sleep(this.msPerGeneration);
		}
	}
	
	public void Clear()
	{
		this.ClearBoard(this.backgroundBoard);
		this.ClearBoard(this.publicBoard);
	
		this.switchBuffer();
		
		this.nrOfGeneration = 0;

		if (this.NextGenerationCompleted != null)
			this.NextGenerationCompleted();
	}
	
	public void SetStartPattern(StartPatterns sp)
	{
		this.ClearBoard(this.backgroundBoard);
		this.ClearBoard(this.publicBoard);
		
		// detect the middle of the board
		int startX = (this.size.Width - 4) / 2;
		int startY = (this.size.Height - 4) / 2;
		
		// copy the pattern to the detected place
		bool [,] pattern = GOLStartPattern.GetStartPattern(sp);
		for (int y=startY; y<startY+4; y++)
			for (int x=startX; x<startX+4; x++)
				this.backgroundBoard[y, x].IsAlive = pattern[y-startY, x-startX];
		
		this.switchBuffer();

		if (this.NextGenerationCompleted != null)
			this.NextGenerationCompleted();
	}
	
	private void ClearBoard(Cell [,] board)
	{
		for (int y=0; y<this.size.Height; y++)
			for (int x=0; x<this.size.Width; x++)
				board[y, x].Reset();
	}
	
	private void IncrementNeighbours(int y, int x)
	{
		// take her neighbours
		for (int cellY = y - 1; cellY < y + 2; cellY++)
		{
			// Continue, if the edge of the board is reached
			if (cellY < 0 || cellY >= this.size.Height)
				continue;

			for (int cellX = x - 1; cellX < x + 2; cellX++)
			{
				// Continue, if the edge of the board is reached
				if (cellX < 0 || cellX >= this.size.Width)
					continue;

				// Continue, if the current cell itself is viewed
				if (cellX == x && cellY == y)
					continue;

				// Increase the cell's neighbour's nrOfNeighbours
				this.backgroundBoard[cellY, cellX].NrOfNeighbours++;
			}
		}
	}
	
	public void BuildNextGeneration()
	{
		// Take each living cell and increment it's neighbours' nrOfNeighbours
		
		// Run through Y-direction
		for (int y=0; y<this.size.Height; y++)
		{
			// Run through X-direction
			for (int x=0; x<this.size.Width; x++)
			{
				// regard every living cell
				if (this.publicBoard[y, x].IsAlive)
					this.IncrementNeighbours(y, x);
			}
		}

		// Take each cell and apply the game's rules

		// Run through Y-direction
		for (int y = 0; y < this.size.Height; y++)
		{
			// Run through X-direction
			for (int x = 0; x < this.size.Width; x++)
			{
				// regard every cell
				if (this.backgroundBoard[y, x].NrOfNeighbours < 2)
					this.backgroundBoard[y, x].IsAlive = false;

				if (this.backgroundBoard[y, x].NrOfNeighbours == 2 && this.publicBoard[y, x].IsAlive)
					this.backgroundBoard[y, x].IsAlive = true;

				if (this.backgroundBoard[y, x].NrOfNeighbours > 3)
					this.backgroundBoard[y, x].IsAlive = false;

				if (this.backgroundBoard[y, x].NrOfNeighbours == 3)
					this.backgroundBoard[y, x].IsAlive = true;
			}
		}

		this.switchBuffer();
		this.ClearBoard(this.backgroundBoard);
		
		this.nrOfGeneration++;
	
		if (this.NextGenerationCompleted != null)
			this.NextGenerationCompleted();
	}
	
	public void TouchCell(Point location)
	{
		for (int y = 0; y < this.size.Height; y++)
			for (int x = 0; x < this.size.Width; x++)
				if (publicBoard[y, x].IsAlive)
					this.backgroundBoard[y, x].IsAlive = true;

		this.backgroundBoard[location.Y, location.X].IsAlive = !this.backgroundBoard[location.Y, location.X].IsAlive;

		this.switchBuffer();
		this.ClearBoard(this.backgroundBoard);

		if (this.NextGenerationCompleted != null)
			this.NextGenerationCompleted();
	}
}